// ---------------------------------------------------------------------------
// sunrise / sunset time library

#pragma GCC diagnostic ignored "-Wwrite-strings"            // always ignore warning about using strings as constants
//#pragma GCC diagnostic ignored "-Wunused-variable"        // ignore unused variables until near end of development cycle, then comment this

#include <math.h>

#include "suntimelib.h"

// ----------------------------------------------------------------------------
// Private Constants
// ----------------------------------------------------------------------------

enum sunRiseModes {
    sunrising                           = 0,
    sunsetting
    };

enum zenithModes {
    zenithOfficial                      = 0,
    zenithCivil,
    zenithNautical,
    zenithAstronomical
    };

const double            zenithDeg[]     = {90.833333, 96.0, 102.0, 108.0};

// ---------------------------------------------------------------------------
// Private Forward declarations

time_t sunRiseSet(
    int                 sunMode,
    time_t              date,
    double              latitude,
    double              longitude,
    int                 zenithMode,
    double              gmtOffset
    );

//----------------------------------------------------------------------------
// The algorith uses degrees, so the following functions are trig functions wrappers using degrees rather than radions

double acosDeg(
    double              n
    ) {

return (180 / M_PI) * acos(n);
} // acosDeg

//----------------------------------------------------------------------------

double asinDeg(
    double              n
    ) {

return (180 / M_PI) * asin(n);

} // asinDeg

//----------------------------------------------------------------------------

double atanDeg(
    double              n
    ) {

return (180 / M_PI) * atan(n);

} // atanDeg

//----------------------------------------------------------------------------

double cosDeg(
    double              angle
    ) {

return cos((M_PI / 180) * angle);

} // cosDeg

//----------------------------------------------------------------------------

double sinDeg(
    double              angle
    ) {

return sin((M_PI / 180) * angle);

} // sinDeg

//----------------------------------------------------------------------------

double tanDeg(
    double              angle
    ) {

return tan((M_PI / 180) * angle);

} // tanDeg

//----------------------------------------------------------------------------
//
// Given date, lat, long, and GMT offset, returns the official sunrise time

time_t sunRise(
    time_t              date,
    double              latitude,
    double              longitude,
    double              gmtOffset
    ) {

return sunRiseSet(sunrising, date, latitude, longitude, zenithOfficial, gmtOffset);

} // sunRise

//----------------------------------------------------------------------------
//
// Given mode (sunrise/sunset), date, lat, long, zenithMode, and GMT offset,
// return the time the sunsets or rises.

// This code uses the algorith found at http://web.archive.org/web/20160315083337/http://williams.best.vwh.net/sunrise_sunset_algorithm.htm

time_t sunRiseSet(
    int                 sunMode,
    time_t              date,
    double              latitude,
    double              longitude,
    int                 zenithMode,
    double              gmtOffset) {

double                  cosDec;                             // COS of sun's declination
double                  cosH;                               // sun's local hour angle
int                     day;                                // day to use for computation
double                  h;
double                  l;                                  // sun's true longitude
double                  lQuadrant;                          // longitude quadrant
double                  lngHour;                            // hour of the longitude
double                  m;                                  // sun's mean anamoly
int                     month;                              // month to use for computation
int                     n;                                  // used to calc julian date day
int                     n1;                                 // used to calc julian date day
int                     n2;                                 // used to calc julian date day
int                     n3;                                 // used to calc julian date day
double                  ra;                                 // sun's right acension
double                  raQuadrant;                         // right acension quadrant
time_t                  result;                             // final result to be returned to caller
double                  sinDec;                             // SIN of sun's declination
double                  t;                                  // approximate time
double                  tLocal;                             // time in local time
double                  tLocalMean;                         // mean time of sun's rising or setting
struct tm               *tm_info;                           // used to convert time_t <-> struct tm
double                  tUTC;                               // time in UTC
int                     year;                               // year to use for computation

if (sunMode < 0 || sunMode > 1) {
    return 0;
    }
if (latitude < -90 || latitude > 90) {
    return 0;
    }
if (longitude < -180 || latitude > 180) {
    return 0;
    }
if (zenithMode < 0 || zenithMode > 3) {
    return 0;
    }

tm_info     = localtime(&date);
month       = tm_info->tm_mon + 1;
day         = tm_info->tm_mday;
year        = tm_info->tm_year + 1900;

// Calculate day of the year (n)
n1  = floor(275 * month / 9);
n2  = floor((month + 9) / 12);
n3  = (1 + floor((year - 4 * floor(year / 4) + 2) / 3));
n   = n1 - (n2 * n3) + day - 30;

// Convert longitude to hour and calc approximate time
lngHour = longitude / 15.0;

if (sunMode == sunrising)
    t = n + ((6 - lngHour) / 24.0);
else
    t = n + ((18 - lngHour) / 24.0);

// Calc the Sun's Mean anomaly
m = 0.9856 * t - 3.289;

// Calc Sun's true longitude
l = m + (1.916 * sinDeg(m)) + (0.020 * sinDeg(2 * m)) + 282.634;
if (l < 0)
    l = l + 360;
else if (l >= 360)
    l = l - 360;

// Calc the sun's right ascension
ra = atanDeg(0.91764 * tanDeg(l));
if (ra < 0)
    ra = ra + 360;
else if (ra > 360)
    ra = ra - 360;

// Right ascension needs to be in the same quadrant as L
lQuadrant = (int)(l / 90) * 90;
raQuadrant = (int)(ra / 90) * 90;
ra = ra + (lQuadrant - raQuadrant);

// Convert Right ascension into hours
ra = ra / 15.0;

// Calc the Sun's declination
sinDec = 0.39782 * sinDeg(l);
cosDec = cosDeg(asinDeg(sinDec));

// Calc the Sun's local hour angle
cosH = (cosDeg(zenithDeg[zenithMode]) - sinDec * sinDeg(latitude)) / (cosDec * cosDeg(latitude));

// Are we at lat/long where sun doesn't rise or set on this day? If so, return error
if (cosH > 1 || cosH < -1)
    return 0;

// Finish calculating H and convert into hours
if (sunMode == sunrising)
    h = 360 - acosDeg(cosH);
else
    h = acosDeg(cosH);
h = h / 15.0;

// Calculate local mean time of rising/setting
tLocalMean = h + ra - (0.06571 * t) - 6.622;

// adjust back to UTC
tUTC = tLocalMean - lngHour;
if (tUTC < 0)
    tUTC = tUTC + 24;
else if (tUTC >= 24)
    tUTC = tUTC - 24;

// adjust back to local time
tLocal = tUTC + gmtOffset;

if (tLocal < 0)
    tLocal = tLocal + 24;
else if (tLocal >= 24)
    tLocal = tLocal - 24;

// if minutes would round up to 60, go to next hour:
if (round((tLocal-floor(tLocal))*60) == 60)
    tLocal = floor(tLocal) + 1;

tm_info->tm_hour = floor(tLocal);
tm_info->tm_min  = round((tLocal-floor(tLocal))*60);
tm_info->tm_sec  = 0;

result = mktime(tm_info);

return result;

} // sunRiseSet

//----------------------------------------------------------------------------
//
// Given date, lat, long, and GMT offset, returns the official sunrise time

time_t sunSet(
    time_t              date,
    double              latitude,
    double              longitude,
    double              gmtOffset
    ) {

return sunRiseSet(sunsetting, date, latitude, longitude, zenithOfficial, gmtOffset);

} // sunSet

